#include "Game.hpp"
#include "Problem.hpp"
#include "Ansi.hpp"
#include "Transform.hpp"
#include "Move.hpp"
#include <limits>
#include <algorithm>

using std::string;
using std::vector;

Game::Game() :
    m_done(false),
    m_sourceLength(0),
    m_spawnCount(0),
    m_score(0),
    m_linesOld(0)
{
}

void Game::init(const Problem& problem,
                const vector<string>& phrases,
                uint32_t seed)
{
    m_phrases = phrases;
    m_chars.clear();
    m_phraseCounts.clear();
    m_phraseCounts.resize(m_phrases.size());

    m_grid.resize(problem.m_width, problem.m_height);

    for (auto cell : problem.m_filled)
    {
        m_grid(cell.m_x, cell.m_y) = 1;
    }

    m_units = problem.m_units;
    for (auto& unit : m_units)
    {
        int32_t minX = std::numeric_limits<int32_t>::max();
        int32_t maxX = std::numeric_limits<int32_t>::min();
        int32_t minY = std::numeric_limits<int32_t>::max();
        for (auto member : unit.m_members)
        {
            minX = std::min(minX, member.m_x);
            maxX = std::max(maxX, member.m_x);
            minY = std::min(minY, member.m_y);
        }

        int32_t dmidX = minX + maxX;
        int32_t dmidXGrid = (int32_t)m_grid.getWidth() - 1;
        int32_t offsetX = (dmidXGrid - dmidX) >> 1;
        int32_t offsetY = -minY;

        unit.m_pivot.m_x += offsetX;
        unit.m_pivot.m_y += offsetY;
        for (auto& member : unit.m_members)
        {
            member.m_x += offsetX;
            member.m_y += offsetY;
        }

        std::sort(unit.m_members.begin(), unit.m_members.end());
    }

    m_sourceLength = problem.m_sourceLength;
    m_source = Source(seed);
    m_spawnCount = 0;
    m_score = 0;
    m_linesOld = 0;
    spawn();
}

bool Game::isValid(const Unit& unit) const
{
    for (auto member : unit.m_members)
    {
        if (member.m_x < 0 ||
            member.m_x >= (int32_t)m_grid.getWidth() ||
            member.m_y < 0 ||
            member.m_y >= (int32_t)m_grid.getHeight() ||
            m_grid(member.m_x, member.m_y) != 0)
        {
            return false;
        }
    }

    return true;
}

void Game::spawn()
{
    if (m_spawnCount >= m_sourceLength)
    {
        m_done = true;
        return;
    }
    m_spawnCount++;

    uint32_t iUnit = m_source.next() % (uint32_t)m_units.size();
    if (isValid(m_units[iUnit]))
    {
        m_curUnit = m_units[iUnit];
    }
    else
    {
        m_done = true;
    }
}

namespace
{
    inline char lc(char ch)
    {
        return (ch >= 'A' && ch <= 'Z') ? (ch - 'A' + 'a') : ch;
    }
}

void Game::step(char ch)
{
    Command command = Command::E;
    bool bIllegal = false;
    if (!charToCommand(ch, &command, &bIllegal))
    {
        if (bIllegal)
        {
            m_done = true;
            m_score = 0;
        }
        return;
    }

    m_chars.push_back(ch);

    if (m_done)
    {
        m_score = 0;
        return;
    }

    for (size_t iPhrase = 0; iPhrase < m_phrases.size(); iPhrase++)
    {
        auto& phrase = m_phrases[iPhrase];
        if (phrase.size() <= m_chars.size())
        {
            size_t start = m_chars.size() - phrase.size();
            size_t i = 0;
            for (i = 0; i < phrase.size(); i++)
            {
                if (lc(phrase[i]) != lc(m_chars[start + i]))
                {
                    break;
                }
            }
            if (i == phrase.size())
            {
                m_score += 2 * phrase.size();
                if (m_phraseCounts[iPhrase] == 0)
                {
                    m_score += 300;
                }
                m_phraseCounts[iPhrase]++;
            }
        }
    }

    Unit unit = m_curUnit;
    transformUnit(command, &unit);
    if (isValid(unit))
    {
        m_curUnit = std::move(unit);
    }
    else
    {
        finishMove();
    }
}

void Game::step(Command command)
{
    step(commandToChar(command));
}

void Game::move(const Move& move)
{
    m_curUnit.m_members = move.m_members;
    finishMove();
}

void Game::finishMove()
{
    uint64_t size = m_curUnit.m_members.size();
    uint64_t lines = 0;
    vector<int32_t> rows;
    for (auto member : m_curUnit.m_members)
    {
        m_grid(member.m_x, member.m_y) = 1;
        rows.push_back(member.m_y);
    }
    std::sort(rows.begin(), rows.end());
    rows.erase(std::unique(rows.begin(), rows.end()), rows.end());
    for (auto row : rows)
    {
        bool fullRow = true;
        for (uint32_t x = 0; x < m_grid.getWidth(); x++)
        {
            if (m_grid(x, row) == 0)
            {
                fullRow = false;
                break;
            }
        }
        if (fullRow)
        {
            for (uint32_t y = row; y > 0; y--)
            {
                for (uint32_t x = 0; x < m_grid.getWidth(); x++)
                {
                    m_grid(x, y) = m_grid(x, y - 1);
                }
            }
            for (uint32_t x = 0; x < m_grid.getWidth(); x++)
            {
                m_grid(x, 0) = 0;
            }
            lines++;
        }
    }
    uint64_t points = size + 100 * (((1 + lines) * lines) / 2);
    uint64_t linesBonus =
        (m_linesOld > 1) ?
        ((m_linesOld - 1) * points) / 10 :
        0;
    uint64_t moveScore = points + linesBonus;
    m_score += moveScore;
    m_linesOld = lines;
    spawn();
}

void printGame(Ansi& ansi, const Game& game)
{
    for (size_t y = 0; y <= game.m_grid.getHeight(); y++)
    {
        ansi.changeColor(LightGray, DarkGray, true);
        printf((y & 1) != 0 ? "  " : " ");
        for (size_t x = 0; x < game.m_grid.getWidth(); x++)
        {
            if (y == game.m_grid.getHeight())
            {
                ansi.changeColor(LightGray, DarkGray, true);
                printf("  ");
            }
            else
            {
                bool bPivot =
                    !game.m_done &&
                    (game.m_curUnit.m_pivot.m_x == (int32_t)x) &&
                    (game.m_curUnit.m_pivot.m_y == (int32_t)y);
                bool bUnit = !game.m_done &&
                    (std::find(game.m_curUnit.m_members.begin(),
                              game.m_curUnit.m_members.end(),
                              Cell((int32_t)x, (int32_t)y)) !=
                     game.m_curUnit.m_members.end());
                bool bFilled = (game.m_grid(x, y) != 0);
                if (bFilled)
                {
                    ansi.changeColor(Black, Yellow, false);
                }
                else if (bUnit)
                {
                    ansi.changeColor(Black, LightCyan, false);
                }
                else
                {
                    ansi.changeColor(DarkGray, Black, false);
                }
                printf(bPivot ? "()" : "<>");
            }
        }
        ansi.changeColor(LightGray, DarkGray, true);
        printf((y & 1) != 0 ? " " : "  ");
        ansi.changeColor(LightGray, Black, false);
        printf("\n");
    }
    printf("Score: %" PRIu64 "\n", game.m_score);
}
