#include "Input.hpp"

#if PLATFORM_WINDOWS
#include <windows.h>
#else
#include <termios.h>
#include <unistd.h>
#endif

using std::string;

class InputImpl
{
public:
#if PLATFORM_WINDOWS
    HANDLE m_hStdIn;
    DWORD m_mode;
    uint32_t m_repeatCount;
    Key m_repeatKey;
#else
    bool m_esc;
    bool m_escLeftBracket;
    bool m_escO;
    bool m_escEsc;
    bool m_escEscLeftBracket;
    bool m_escEscO;
    string m_digits;
    struct termios m_oldt;
#endif
};

Input::Input() :
    m_pImpl(new InputImpl)
{
#if PLATFORM_WINDOWS
    m_pImpl->m_hStdIn = GetStdHandle(STD_INPUT_HANDLE);
    m_pImpl->m_mode = 0;
    GetConsoleMode(m_pImpl->m_hStdIn, &m_pImpl->m_mode);
    SetConsoleMode(m_pImpl->m_hStdIn, ENABLE_EXTENDED_FLAGS | ENABLE_PROCESSED_INPUT);
    m_pImpl->m_repeatCount = 0;
    m_pImpl->m_repeatKey = Key::Invalid;
#else
    m_pImpl->m_esc = false;
    m_pImpl->m_escLeftBracket = false;
    m_pImpl->m_escO = false;
    m_pImpl->m_escEsc = false;
    m_pImpl->m_escEscLeftBracket = false;
    m_pImpl->m_escEscO = false;

    int stdin_fd = fileno(stdin);
    tcgetattr(stdin_fd, &m_pImpl->m_oldt);
    struct termios newt;
    newt = m_pImpl->m_oldt;
    newt.c_lflag &= ~(ICANON | ECHO);
    newt.c_iflag &= ~(INLCR | IGNCR | ICRNL | IXON | IXOFF);
    tcsetattr(stdin_fd, TCSANOW, &newt);
#endif
}

Input::~Input()
{
#if PLATFORM_WINDOWS
    SetConsoleMode(m_pImpl->m_hStdIn, m_pImpl->m_mode);
#else
    int stdin_fd = fileno(stdin);
    m_pImpl->m_oldt.c_lflag |= (ICANON | ECHO);
    tcsetattr(stdin_fd, TCSANOW, &m_pImpl->m_oldt);
#endif
}

#if !PLATFORM_WINDOWS
namespace
{
    int getChar(bool blocking)
    {
        struct termios oldt, newt;
        int stdin_fd = fileno(stdin);
        tcgetattr(stdin_fd, &oldt);
        newt = oldt;
        newt.c_cc[VMIN] = blocking ? 1 : 0;
        newt.c_cc[VTIME] = 0;
        tcsetattr(stdin_fd, TCSANOW, &newt);
        int ch = getchar();
        tcsetattr(stdin_fd, TCSANOW, &oldt);
        return ch;
    }
}
#endif

bool Input::getKey(bool blocking, Key* pKey)
{
    Key dummyKey = Key::Invalid;
    if (!pKey)
    {
        pKey = &dummyKey;
    }

#if PLATFORM_WINDOWS
    while (true)
    {
        if (m_pImpl->m_repeatCount > 0)
        {
            m_pImpl->m_repeatCount--;
            *pKey = m_pImpl->m_repeatKey;
            return true;
        }

        INPUT_RECORD inputRecord;
        DWORD numRead = 0;

        if (!blocking)
        {
            if (!PeekConsoleInput(m_pImpl->m_hStdIn, &inputRecord, 1, &numRead))
            {
                return false;
            }
            if (numRead == 0)
            {
                return false;
            }
        }

        if (!ReadConsoleInput(m_pImpl->m_hStdIn, &inputRecord, 1, &numRead))
        {
            return false;
        }
        if (numRead == 0)
        {
            return false;
        }
        if (inputRecord.EventType != KEY_EVENT)
        {
            continue;
        }
        if (!inputRecord.Event.KeyEvent.bKeyDown)
        {
            continue;
        }
        DWORD keyState = inputRecord.Event.KeyEvent.dwControlKeyState;
        bool bShift = ((keyState & SHIFT_PRESSED) != 0);
        bool bCtrl = ((keyState & (LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED)) != 0);
        bool bAlt = ((keyState & (LEFT_ALT_PRESSED | RIGHT_ALT_PRESSED)) != 0);
        bool bCaps = ((keyState & CAPSLOCK_ON) != 0);
        Key key = Key::Invalid;

        switch (inputRecord.Event.KeyEvent.wVirtualKeyCode)
        {
        case VK_BACK: key = Key::Backspace; break;
        case VK_TAB: key = Key::Tab; break;
        case VK_CLEAR: key = Key::Center; break;
        case VK_RETURN: key = Key::Enter; break;
        case VK_ESCAPE: key = Key::Esc; break;
        case VK_SPACE: key = Key::Space; break;
        case VK_PRIOR: key = Key::PgUp; break;
        case VK_NEXT: key = Key::PgDn; break;
        case VK_END: key = Key::End; break;
        case VK_HOME: key = Key::Home; break;
        case VK_LEFT: key = Key::Left; break;
        case VK_UP: key = Key::Up; break;
        case VK_RIGHT: key = Key::Right; break;
        case VK_DOWN: key = Key::Down; break;
        case VK_INSERT: key = Key::Insert; break;
        case VK_DELETE: key = Key::Delete; break;
        case '0': key = Key::Key0; break;
        case '1': key = Key::Key1; break;
        case '2': key = Key::Key2; break;
        case '3': key = Key::Key3; break;
        case '4': key = Key::Key4; break;
        case '5': key = Key::Key5; break;
        case '6': key = Key::Key6; break;
        case '7': key = Key::Key7; break;
        case '8': key = Key::Key8; break;
        case '9': key = Key::Key9; break;
        case 'A': key = Key::A; break;
        case 'B': key = Key::B; break;
        case 'C': key = Key::C; break;
        case 'D': key = Key::D; break;
        case 'E': key = Key::E; break;
        case 'F': key = Key::F; break;
        case 'G': key = Key::G; break;
        case 'H': key = Key::H; break;
        case 'I': key = Key::I; break;
        case 'J': key = Key::J; break;
        case 'K': key = Key::K; break;
        case 'L': key = Key::L; break;
        case 'M': key = Key::M; break;
        case 'N': key = Key::N; break;
        case 'O': key = Key::O; break;
        case 'P': key = Key::P; break;
        case 'Q': key = Key::Q; break;
        case 'R': key = Key::R; break;
        case 'S': key = Key::S; break;
        case 'T': key = Key::T; break;
        case 'U': key = Key::U; break;
        case 'V': key = Key::V; break;
        case 'W': key = Key::W; break;
        case 'X': key = Key::X; break;
        case 'Y': key = Key::Y; break;
        case 'Z': key = Key::Z; break;
        case VK_F1: key = Key::F1; break;
        case VK_F2: key = Key::F2; break;
        case VK_F3: key = Key::F3; break;
        case VK_F4: key = Key::F4; break;
        case VK_F5: key = Key::F5; break;
        case VK_F6: key = Key::F6; break;
        case VK_F7: key = Key::F7; break;
        case VK_F8: key = Key::F8; break;
        case VK_F9: key = Key::F9; break;
        case VK_F10: key = Key::F10; break;
        case VK_F11: key = Key::F11; break;
        case VK_F12: key = Key::F12; break;
        case VK_OEM_1: key = Key::Semicolon; break;
        case VK_OEM_PLUS: key = Key::Equals; break;
        case VK_OEM_COMMA: key = Key::Comma; break;
        case VK_OEM_MINUS: key = Key::Minus; break;
        case VK_OEM_PERIOD: key = Key::Period; break;
        case VK_OEM_2: key = Key::Slash; break;
        case VK_OEM_3: key = Key::Backtick; break;
        case VK_OEM_4: key = Key::LeftBracket; break;
        case VK_OEM_5: key = Key::Backslash; break;
        case VK_OEM_6: key = Key::RightBracket; break;
        case VK_OEM_7: key = Key::Apostrophe; break;
        default: continue;
        }

        bool bAdjustedShift = bShift;
        if (bCaps && (key >= Key::A && key <= Key::Z))
        {
            bAdjustedShift = !bAdjustedShift;
        }

        if (bAdjustedShift)
        {
            key = (Key)((uint32_t)key | (uint32_t)Key::Shift);
        }
        if (bCtrl)
        {
            key = (Key)((uint32_t)key | (uint32_t)Key::Ctrl);
        }
        if (bAlt)
        {
            key = (Key)((uint32_t)key | (uint32_t)Key::Alt);
        }

        if (inputRecord.Event.KeyEvent.wRepeatCount > 1)
        {
            m_pImpl->m_repeatCount = inputRecord.Event.KeyEvent.wRepeatCount - 1;
            m_pImpl->m_repeatKey = key;
        }

        *pKey = key;
        return true;
    }

    return false;
#else
    while (true)
    {
        int ch = getChar((m_pImpl->m_esc ||
                          m_pImpl->m_escEsc ||
                          m_pImpl->m_escLeftBracket ||
                          m_pImpl->m_escO) ? false : blocking);
        if (ch == EOF)
        {
            if (m_pImpl->m_esc)
            {
                m_pImpl->m_esc = false;
                *pKey = Key::Esc;
                return true;
            }
            else if (m_pImpl->m_escEsc)
            {
                m_pImpl->m_escEsc = false;
                *pKey = Key::AltEsc;
                return true;
            }
            else if (m_pImpl->m_escLeftBracket)
            {
                m_pImpl->m_escLeftBracket = false;
                *pKey = Key::AltLeftBracket;
                return true;
            }
            else if (m_pImpl->m_escO)
            {
                m_pImpl->m_escO = false;
                *pKey = Key::AltShiftO;
                return true;
            }
            return false;
        }

        Key key = Key::Invalid;

        if (m_pImpl->m_escLeftBracket ||
            m_pImpl->m_escEscLeftBracket)
        {
            switch (ch)
            {
            case 'A': key = Key::Up; break;
            case 'B': key = Key::Down; break;
            case 'C': key = Key::Right; break;
            case 'D': key = Key::Left; break;
            case 'G': key = Key::Center; break;
            case 'Z': key = Key::ShiftTab; break;
            case '0':
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
            case '8':
            case '9':
                m_pImpl->m_digits.push_back((char)ch);
                continue;
            case '~':
                if (strcmp(m_pImpl->m_digits.c_str(), "1") == 0) { key = Key::Home; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "2") == 0) { key = Key::Insert; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "3") == 0) { key = Key::Delete; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "4") == 0) { key = Key::End; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "5") == 0) { key = Key::PgUp; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "6") == 0) { key = Key::PgDn; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "11") == 0) { key = Key::F1; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "12") == 0) { key = Key::F2; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "13") == 0) { key = Key::F3; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "14") == 0) { key = Key::F4; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "15") == 0) { key = Key::F5; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "17") == 0) { key = Key::F6; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "18") == 0) { key = Key::F7; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "19") == 0) { key = Key::F8; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "20") == 0) { key = Key::F9; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "21") == 0) { key = Key::F10; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "23") == 0) { key = Key::F11; break; }
                if (strcmp(m_pImpl->m_digits.c_str(), "24") == 0) { key = Key::F12; break; }
            default:
                m_pImpl->m_escLeftBracket = false;
                m_pImpl->m_escEscLeftBracket = false;
                m_pImpl->m_digits.clear();
                continue;
            }
            if (m_pImpl->m_escEscLeftBracket)
            {
                key = (Key)((uint32_t)key | (uint32_t)Key::Alt);
            }
            m_pImpl->m_escLeftBracket = false;
            m_pImpl->m_escEscLeftBracket = false;
            m_pImpl->m_digits.clear();
        }
        else if (m_pImpl->m_escO ||
                 m_pImpl->m_escEscO)
        {
            switch (ch) {
            case 'A': key = Key::CtrlUp; break;
            case 'B': key = Key::CtrlDown; break;
            case 'C': key = Key::CtrlRight; break;
            case 'D': key = Key::CtrlLeft; break;
            case 'G': key = Key::CtrlCenter; break;
            default:
                m_pImpl->m_escO = false;
                m_pImpl->m_escEscO = false;
                continue;
            }
            if (m_pImpl->m_escEscO)
            {
                key = (Key)((uint32_t)key | (uint32_t)Key::Alt);
            }
            m_pImpl->m_escO = false;
            m_pImpl->m_escEscO = false;
        }
        else if (m_pImpl->m_escEsc)
        {
            m_pImpl->m_escEsc = false;
            switch (ch)
            {
            case '[': m_pImpl->m_escEscLeftBracket = true; continue;
            case 'O': m_pImpl->m_escEscO = true; continue;
            default: continue;
            }
        }
        else
        {
            switch (ch)
            {
            case 0x00: key = Key::CtrlSpace; break;
            case 0x01: key = Key::CtrlA; break;
            case 0x02: key = Key::CtrlB; break;
            case 0x03: key = Key::CtrlC; break;
            case 0x04: key = Key::CtrlD; break;
            case 0x05: key = Key::CtrlE; break;
            case 0x06: key = Key::CtrlF; break;
            case 0x07: key = Key::CtrlG; break;
            case 0x08: key = Key::CtrlH; break;
            case 0x09: key = Key::Tab; break;
            case 0x0a: key = Key::CtrlJ; break;
            case 0x0b: key = Key::CtrlK; break;
            case 0x0c: key = Key::CtrlL; break;
            case 0x0d: key = Key::Enter; break;
            case 0x0e: key = Key::CtrlN; break;
            case 0x0f: key = Key::CtrlO; break;
            case 0x10: key = Key::CtrlP; break;
            case 0x11: key = Key::CtrlQ; break;
            case 0x12: key = Key::CtrlR; break;
            case 0x13: key = Key::CtrlS; break;
            case 0x14: key = Key::CtrlT; break;
            case 0x15: key = Key::CtrlU; break;
            case 0x16: key = Key::CtrlV; break;
            case 0x17: key = Key::CtrlW; break;
            case 0x18: key = Key::CtrlX; break;
            case 0x19: key = Key::CtrlY; break;
            case 0x1a: key = Key::CtrlZ; break;
            case 0x1b:
                if (m_pImpl->m_esc)
                {
                    m_pImpl->m_esc = false;
                    m_pImpl->m_escEsc = true;
                }
                else
                {
                    m_pImpl->m_esc = true;
                }
                continue;
            case 0x1c: key = Key::CtrlBackslash; break;
            case 0x1d: key = Key::CtrlRightBracket; break;
            case 0x1e: key = Key::Ctrl6; break;
            case 0x1f: key = Key::CtrlMinus; break;
            case 0x7f: key = Key::Backspace; break;
            case ' ': key = Key::Space; break;
            case '0': key = Key::Key0; break;
            case '1': key = Key::Key1; break;
            case '2': key = Key::Key2; break;
            case '3': key = Key::Key3; break;
            case '4': key = Key::Key4; break;
            case '5': key = Key::Key5; break;
            case '6': key = Key::Key6; break;
            case '7': key = Key::Key7; break;
            case '8': key = Key::Key8; break;
            case '9': key = Key::Key9; break;
            case 'a': key = Key::A; break;
            case 'b': key = Key::B; break;
            case 'c': key = Key::C; break;
            case 'd': key = Key::D; break;
            case 'e': key = Key::E; break;
            case 'f': key = Key::F; break;
            case 'g': key = Key::G; break;
            case 'h': key = Key::H; break;
            case 'i': key = Key::I; break;
            case 'j': key = Key::J; break;
            case 'k': key = Key::K; break;
            case 'l': key = Key::L; break;
            case 'm': key = Key::M; break;
            case 'n': key = Key::N; break;
            case 'o': key = Key::O; break;
            case 'p': key = Key::P; break;
            case 'q': key = Key::Q; break;
            case 'r': key = Key::R; break;
            case 's': key = Key::S; break;
            case 't': key = Key::T; break;
            case 'u': key = Key::U; break;
            case 'v': key = Key::V; break;
            case 'w': key = Key::W; break;
            case 'x': key = Key::X; break;
            case 'y': key = Key::Y; break;
            case 'z': key = Key::Z; break;
            case ';': key = Key::Semicolon; break;
            case '=': key = Key::Equals; break;
            case ',': key = Key::Comma; break;
            case '-': key = Key::Minus; break;
            case '.': key = Key::Period; break;
            case '/': key = Key::Slash; break;
            case '`': key = Key::Backtick; break;
            case '[':
                if (m_pImpl->m_esc)
                {
                    m_pImpl->m_esc = false;
                    m_pImpl->m_escLeftBracket = true;
                    continue;
                }
                key = Key::LeftBracket; break;
            case '\\': key = Key::Backslash; break;
            case ']': key = Key::RightBracket; break;
            case '\'': key = Key::Apostrophe; break;
            case '!': key = Key::Exclamation; break;
            case '@': key = Key::At; break;
            case '#': key = Key::Pound; break;
            case '$': key = Key::Dollar; break;
            case '%': key = Key::Percent; break;
            case '^': key = Key::Caret; break;
            case '&': key = Key::Ampersand; break;
            case '*': key = Key::Asterisk; break;
            case '(': key = Key::LeftParen; break;
            case ')': key = Key::RightParen; break;
            case 'A': key = Key::ShiftA; break;
            case 'B': key = Key::ShiftB; break;
            case 'C': key = Key::ShiftC; break;
            case 'D': key = Key::ShiftD; break;
            case 'E': key = Key::ShiftE; break;
            case 'F': key = Key::ShiftF; break;
            case 'G': key = Key::ShiftG; break;
            case 'H': key = Key::ShiftH; break;
            case 'I': key = Key::ShiftI; break;
            case 'J': key = Key::ShiftJ; break;
            case 'K': key = Key::ShiftK; break;
            case 'L': key = Key::ShiftL; break;
            case 'M': key = Key::ShiftM; break;
            case 'N': key = Key::ShiftN; break;
            case 'O':
                if (m_pImpl->m_esc)
                {
                    m_pImpl->m_esc = false;
                    m_pImpl->m_escO = true;
                    continue;
                }
                key = Key::ShiftO; break;
            case 'P': key = Key::ShiftP; break;
            case 'Q': key = Key::ShiftQ; break;
            case 'R': key = Key::ShiftR; break;
            case 'S': key = Key::ShiftS; break;
            case 'T': key = Key::ShiftT; break;
            case 'U': key = Key::ShiftU; break;
            case 'V': key = Key::ShiftV; break;
            case 'W': key = Key::ShiftW; break;
            case 'X': key = Key::ShiftX; break;
            case 'Y': key = Key::ShiftY; break;
            case 'Z': key = Key::ShiftZ; break;
            case ':': key = Key::Colon; break;
            case '+': key = Key::Plus; break;
            case '<': key = Key::Less; break;
            case '_': key = Key::Underscore; break;
            case '>': key = Key::Greater; break;
            case '?': key = Key::Question; break;
            case '~': key = Key::Tilde; break;
            case '{': key = Key::LeftBrace; break;
            case '|': key = Key::VerticalBar; break;
            case '}': key = Key::RightBrace; break;
            case '"': key = Key::Quote; break;
            default:
                m_pImpl->m_esc = false;
                continue;
            }
            if (m_pImpl->m_esc)
            {
                key = (Key)((uint32_t)key | (uint32_t)Key::Alt);
            }
            m_pImpl->m_esc = false;
        }

        *pKey = key;
        return true;
    }

    return false;
#endif
}
