#include "Solver.hpp"
#include "Problem.hpp"
#include "Solution.hpp"
#include "Game.hpp"
#include "Move.hpp"
#include "Moves.hpp"
#include "Utils.hpp"
#include "Transform.hpp"
#include <algorithm>

using std::string;
using std::vector;

bool solve(const Problem& problem,
           const vector<string>& phrases,
           uint32_t seed,
           uint32_t scoreFunc,
           uint32_t scoreSeed,
           uint64_t stopMS,
           Solution* pSolution,
           uint64_t* pScore)
{
    Source scoreSource(scoreSeed);
    scoreSource.next();

    Game game;
    game.init(problem, phrases, seed);

    while (!game.m_done)
    {
        if (stopMS != 0 && getTicksMS() >= stopMS)
        {
            return false;
        }

        MoveFinder moveFinder;

        vector<Move> moves;
        moveFinder.findMoves(game, &moves);

        size_t bestMove = 0;
        vector<int32_t> bestMoveVec;
        uint32_t sameCount = 0;
        for (uint32_t i = 0; i < moves.size(); i++)
        {
            auto& move = moves[i];

            int32_t curMoveYTotal = 0;
            int32_t curMoveYMax = std::numeric_limits<int32_t>::min();
            int32_t curMoveEmpty = 0;
            int32_t curMoveNeighbors = 0;
            vector<int32_t> curMoveVec;
            for (auto& member : move.m_members)
            {
                curMoveYTotal += member.m_y;
                curMoveYMax = std::max(curMoveYMax, member.m_y);

                for (size_t x = 0; x < game.m_grid.getWidth(); x++)
                {
                    if (game.m_grid(x, member.m_y) == 0)
                    {
                        curMoveEmpty++;
                    }
                }

                for (uint32_t iCommand = 0; iCommand < 4; iCommand++)
                {
                    Cell neighbor =
                        transformCell(Command(iCommand),
                                      Cell(),
                                      Cell(member.m_x, member.m_y));
                    if (neighbor.m_x < 0 ||
                        neighbor.m_x >= (int32_t)game.m_grid.getWidth() ||
                        neighbor.m_y < 0 ||
                        neighbor.m_y >= (int32_t)game.m_grid.getHeight() ||
                        game.m_grid(neighbor.m_x, neighbor.m_y) != 0)
                    {
                        curMoveNeighbors++;
                    }
                }
            }

            if (scoreFunc == 0)
            {
                curMoveVec.push_back(curMoveYTotal);
            }
            else if (scoreFunc == 1)
            {
                curMoveVec.push_back(curMoveYMax);
                curMoveVec.push_back(curMoveYTotal);
            }
            else if (scoreFunc == 2)
            {
                curMoveVec.push_back(-curMoveEmpty);
                curMoveVec.push_back(curMoveNeighbors);
            }
            else if (scoreFunc == 3)
            {
                curMoveVec.push_back(curMoveYMax);
                curMoveVec.push_back(curMoveNeighbors);
            }
            else if (scoreFunc == 4)
            {
                curMoveVec.push_back(curMoveYTotal);
                curMoveVec.push_back(curMoveNeighbors);
            }

            bool better = false;
            bool same = false;
            if (i == 0)
            {
                better = true;
            }
            else
            {
                if (std::lexicographical_compare(
                        bestMoveVec.begin(),
                        bestMoveVec.end(),
                        curMoveVec.begin(),
                        curMoveVec.end()))
                {
                    better = true;
                }
                else if (!std::lexicographical_compare(
                             curMoveVec.begin(),
                             curMoveVec.end(),
                             bestMoveVec.begin(),
                             bestMoveVec.end()))
                {
                    same = true;
                }
            }

            if (better)
            {
                bestMove = i;
                bestMoveVec = curMoveVec;
                sameCount = 1;
            }
            else if (same)
            {
                sameCount++;
                if (scoreSource.next() % sameCount == 0)
                {
                    bestMove = i;
                    bestMoveVec = curMoveVec;
                }
            }
        }

        vector<Command> moveCommands;
        moveFinder.getCommands(moves[bestMove], &moveCommands);

        for (auto command : moveCommands)
        {
            char ch = commandToChar(command);
            game.step(ch);
        }
    }

    auto& solution = *pSolution;
    solution.m_problemId = problem.m_id;
    solution.m_seed = seed;
    solution.m_tag = "v4";
    solution.m_chars = game.m_chars;
    *pScore = game.m_score;
    return true;
}
