#include "Common.hpp"
#include "Source.hpp"
#include "Problem.hpp"
#include "Parser.hpp"
#include "Utils.hpp"
#include "Ansi.hpp"
#include "Game.hpp"
#include "Input.hpp"
#include "Moves.hpp"
#include "Move.hpp"

using std::string;
using std::vector;

void usage()
{
    printf("Usage:\n");
    printf("  play <problem.json>\n");
    printf("Keys:\n");
    printf("  Right:      Move east\n");
    printf("  Left:       Move west\n");
    printf("  Down:       Move southeast\n");
    printf("  Ctrl+Down:  Move southwest\n");
    printf("  Ctrl+Left:  Rotate counterclockwise\n");
    printf("  Ctrl+Right: Rotate clockwise\n");
    printf("  Space:      Automatically make one move\n");
    printf("  1:          Automatically perform one command of a move\n");
    printf("  Q or Esc:   Quit\n");
}

int main(int argc, char* argv[])
{
    if (argc != 2)
    {
        usage();
        return 1;
    }

    string strProblemFile(argv[1]);

    string msg;
    Problem problem;
    if (!loadProblem(strProblemFile, &problem, &msg))
    {
        printf("%s\n", msg.c_str());
        return 1;
    }

    Ansi ansi;
    Input input;

    Game game;
    game.init(problem, vector<string>(), problem.m_sourceSeeds[0]);
    ansi.clear();
    printGame(ansi, game);

    vector<Command> commands;
    size_t iCommand = 0;

    while (!game.m_done)
    {
        Key key = Key::Invalid;
        if (!input.getKey(true, &key))
        {
            break;
        }

        Command command = Command::E;
        bool bQuit = false;
        bool bMove = false;
        bool bSingle = false;
        switch (key)
        {
        case Key::Left: command = Command::W; break;
        case Key::Right: command = Command::E; break;
        case Key::CtrlLeft: command = Command::CCW; break;
        case Key::CtrlRight: command = Command::CW; break;
        case Key::Down: command = Command::SE; break;
        case Key::CtrlDown: command = Command::SW; break;
        case Key::Space: bMove = true; break;
        case Key::Key1: bSingle = true; break;
        case Key::Q:
        case Key::Esc:
            bQuit = true; break;
        default: continue;
        }

        if (bQuit)
        {
            break;
        }

        if (bSingle && iCommand < commands.size())
        {
            game.step(commands[iCommand]);
            iCommand++;
        }
        else
        {
            commands.clear();
            iCommand = 0;

            if (bMove || bSingle)
            {
                vector<Move> moves;
                MoveFinder moveFinder;
                moveFinder.findMoves(game, &moves);
                size_t bestMove = 0;
                uint32_t bestMoveScore = 0;
                for (uint32_t i = 0; i < moves.size(); i++)
                {
                    auto& move = moves[i];
                    uint32_t curMoveScore = 0;
                    for (auto& member : move.m_members)
                    {
                        curMoveScore += member.m_y;
                    }
                    if (i == 0 || curMoveScore > bestMoveScore)
                    {
                        bestMove = i;
                        bestMoveScore = curMoveScore;
                    }
                }

                if (bMove)
                {
                    game.move(moves[bestMove]);
                }
                else
                {
                    moveFinder.getCommands(moves[bestMove], &commands);
                    game.step(commands[iCommand]);
                    iCommand++;
                }
            }
            else
            {
                game.step(command);
            }
        }

        ansi.clear();
        printGame(ansi, game);
    }

    return 0;
}
